/* m-6000mtxt.c: The AKIZUKI M-6000M (ES51986) tester serial output reader.

   Copyright 2013 Akinori Furuta<afuruta@m7.dion.ne.jp>.
   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions
   are met:

   1. Redistributions of source code must retain the above copyright notice,
      this list of conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright notice,
      this list of conditions and the following disclaimer in the documentation
      and/or other materials provided with the distribution.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
   AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
   THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
   PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
   CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
   EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
   PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
   OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
   WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
   OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
   ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/
#define _GNU_SOURCE
#include <features.h>
#include <stdint.h>
#include <inttypes.h>
#include <limits.h>
#include <linux/serial.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <fcntl.h>
#include <poll.h>
#include <signal.h>
#include <string.h>
#include <stdlib.h>
#include <termios.h>
#include <errno.h>
#include <time.h>
#include <ctype.h>
#include <stdio.h>

#if (!defined(VERSION_STRING))
#define VERSION_STRING "NotVersioned"
#endif
const char copyright_notice[]="Copyright 2013 Akinori Furuta<afuruta@m7.dion.ne.jp>.\n";
const char version_string[]=VERSION_STRING;

/* Link with -lrt
*/

#define	M6000M_BPS		(19200)
/*! It is most long period in capacitance mode. 
    about 4.5 seconds at 6000uF.
*/
#define	M6K_UPDATE_PERIOD_MS	(5000) /* milli seconds. */

/*! Delta time in milli seconds, from 1st line end to 2nd line start. */
#define	M6K_2ND_LF_MS	(40L)

/*! Timeout retry count. */
#define	COPT_BREAK_RETRY_DEF	(5)

const char	*ProgName="m-6000mtxt";

/*! request help. */
#define	COPT_FLAGS_HELP			(0x0001)
/*! Output invalid reads. */
#define	COPT_FLAGS_INVALID_READ		(0x0002)
/*! Clear(flush) queue.  */
#define	COPT_FLAGS_CLEAR_QUEUE		(0x0008)
/*! Send Break at start. */
#define	COPT_FLAGS_BREAK_AT_START	(0x0010)
/*! Send Break at start. */
#define	COPT_FLAGS_BREAK_AT_TIMEOUT	(0x0020)
/*! Show settings. */
#define	COPT_FLAGS_SHOW_SETTINGS	(0x0040)
/*! Show version. */
#define	COPT_FLAGS_SHOW_VERSION		(0x0080)
/*! Debug lock. */
#define	COPT_FLAGS_DEBUG_LOCK		(0x0100)

/*! Command line
*/
typedef struct {
	unsigned long	Flags;		/*!< Flags. */
	char		*PortPath;	/*!< M-6000M serial port. */
	char		*Format;	/*!< Output format */
	long		DtL1L2;		/*!< Delta time L1-L2. */
	long		Timeout;	/*!< Time to exit. */
	long		BreakRetry;	/*!< Retry break at timeout. */
	char		Blank;		/*!< Blank character. */
} TCommandLine;

TCommandLine	CmdLine={
	.Flags=COPT_FLAGS_BREAK_AT_TIMEOUT,
	.PortPath=NULL,
	.Format="%R %a %e %u",
	.DtL1L2=M6K_2ND_LF_MS,
	.Timeout=(M6K_UPDATE_PERIOD_MS*3)/2,
	.BreakRetry=COPT_BREAK_RETRY_DEF,
	.Blank='#'
};

/*! Program context structure.
*/
typedef	struct {
	int		SignalIntr;
	int		SignalHup;
	int		SignalTerm;
	sighandler_t	PrevHandlerIntr;
	sighandler_t	PrevHandlerHup;
	sighandler_t	PrevHandlerTerm;
} TProgramContext;

/*! Program context.
*/
TProgramContext	ProgContext={
	.SignalIntr=0,
	.SignalHup=0,
	.SignalTerm=0,
	.PrevHandlerIntr=NULL,
	.PrevHandlerHup=NULL,
	.PrevHandlerTerm=NULL
};

/*! Interrupt handler.
*/
void SignalHandlerIntr(int sig)
{	ProgContext.SignalIntr=1;
	fprintf(stderr,"%s: Interrupted.\n", ProgName);
	return;
}

/*! Hangup handler.
*/
void SignalHandlerHup(int sig)
{	ProgContext.SignalHup=1;
	fprintf(stderr,"%s: Hangup.\n", ProgName);
	return;
}

/*! Terminate handler.
*/
void SignalHandlerTerm(int sig)
{	ProgContext.SignalTerm=1;
	fprintf(stderr,"%s: Terminated.\n", ProgName);
	return;
}

/*! Trap signals.
*/
int HandleSignals(void)
{	int		ret;
	sighandler_t	prev;

	ret=1;
	prev=signal(SIGINT,SignalHandlerIntr);
	if (prev==SIG_ERR) {
		fprintf(stderr,"%s: Can not set handler SIGINT. %s(%d).\n"
			,ProgName
			,strerror(errno),errno
		);
		 ret=0;
	}
	ProgContext.PrevHandlerIntr=prev;

	prev=signal(SIGHUP,SignalHandlerHup);
	if (prev==SIG_ERR) {
		fprintf(stderr,"%s: Can not set handler SIGHUP. %s(%d).\n"
			,ProgName
			,strerror(errno),errno
		);
		ret=0;
	}
	ProgContext.PrevHandlerHup=prev;

	prev=signal(SIGTERM,SignalHandlerTerm);
	if (prev==SIG_ERR) {
		fprintf(stderr,"%s: Can not set handler SIGTERM. %s(%d).\n"
			,ProgName
			,strerror(errno),errno
			);
		ret=0;
	}
	ProgContext.PrevHandlerTerm=prev;
	return(ret);
}

/*! Untrap signals.
*/
void UnHandleSignals(void)
{	sighandler_t	prev;

	prev=ProgContext.PrevHandlerIntr;
	if ((prev!=SIG_ERR) && (prev!=NULL)) {
		signal(SIGINT,prev);
	}

	prev=ProgContext.PrevHandlerHup;
	if ((prev!=SIG_ERR) && (prev!=NULL)) {
		signal(SIGHUP,prev);
	}

	prev=ProgContext.PrevHandlerTerm;
	if ((prev!=SIG_ERR) && (prev!=NULL)) {
		signal(SIGTERM,prev);
	}
	return;
}


/*! Skip Space chars.
    @param p points string to skip space characters.
    @return char* points non space character.
*/
char *StrSkipSpace(char *p)
{	char	c;

	if (!p) {
		return p;
	}
	while ((c=*p)!=0) {
		if (!isspace(c)) {
			break;
		}
		p++;
	}
	return(p);
}

/*! Cont word chars.
*/
size_t StrNumCharsWordSpan(char *p)
{	char	*p0;
	char	c;

	if (!p) {
		return 0;
	}
	p0=p;
	while ((c=*p)!=0) {
		if (!isalnum(c)) {
			break;
		}
		p++;
	}
	return p-p0 /* pointer sub. */;
}

/*! ReDefile type TimeSpec as struct timespec */
typedef	struct timespec TimeSpec;

/*! Get CLOCK_REALTIME.
    @param p points TimeSpec to store realtime.
    @return int !=0: Success, 0: Failed.
*/
static int TimeSpecGetRealTime(TimeSpec *p)
{	if (clock_gettime(CLOCK_REALTIME,p)==0) {
		return 1 /* True, Success. */;
	}
	return 0 /* False, Failed. */;
}

#define	TIMESPEC_SPRINT_MAX	(sizeof(unsigned long long)*3+1+sizeof(unsigned long)*3+1)

/*! sprint TimeSpec
    @param p points TimeSpec to sprint.
    @param buf buffer to sprint.
    @return int printed length, not included NUL terminator.
*/
static int TimeSpecSPrint(TimeSpec *p, char *buf)
{	return	sprintf(buf,"%llu.%.9lu"
			,(unsigned long long)(p->tv_sec)
			,(unsigned long)(p->tv_nsec)
		);
}

/*! Sub two TimeSpec(s), *y=*a-*b.
    @param y points TimeSpec structure to store *a-*b
    @param a points TimeSpec structure.
    @param b points TimeSpec structure.
    @return struct TimeSpec pointer equal to y.
    @note *a means end time, *b means start time.
*/
TimeSpec *TimeSpecSub(TimeSpec *y, const TimeSpec *a, const TimeSpec *b)
{	time_t	sec;
	long	nsec;

	sec= a->tv_sec- b->tv_sec;
	nsec=a->tv_nsec-b->tv_nsec;
	if (nsec<0) {
		/* Borrow */
		sec--;
		nsec+=1000L*1000L*1000L;
	}
	y->tv_sec=sec;
	y->tv_nsec=nsec;
	return(y);
}

/*! Compare two TimeSpec(s), int=cmp(*a,*b).
    @param a points TimeSpec structure.
    @param b points TimeSpec structure.
    @return int <0: *a<*b, ==0: *a==*b, >0 *a>*b.
*/
int TimeSpecCmp(const TimeSpec *a, const TimeSpec *b)
{	if (a->tv_sec<b->tv_sec) {
		return -1;
	}
	if (a->tv_sec>b->tv_sec) {
		return 1;
	}
	if (a->tv_nsec<b->tv_nsec) {
		return -1;
	}
	if (a->tv_nsec>b->tv_nsec) {
		return 1;
	}
	return 0;
}

/*! Zero TimeSpec.
    @param y points TimeSpec structure.
*/
static void TimeSpecZero(TimeSpec *y)
{	if (!y) {
		return;
	}
	y->tv_sec=0;
	y->tv_nsec=0;
}

/*! Add two TimeSpec(s), *y=*a+*b.
    @param y points TimeSpec structure to store *a+*b
    @param a points TimeSpec structure.
    @param b points TimeSpec structure.
    @return struct TimeSpec pointer equal to y.
*/
TimeSpec *TimeSpecAdd(TimeSpec *y, const TimeSpec *a, const TimeSpec *b)
{	time_t	sec;
	long	nsec;

	sec=  a->tv_sec+ b->tv_sec;
	nsec= a->tv_nsec+b->tv_nsec;

	if (nsec>=1000L*1000L*1000L) {
		/* Carry. */
		sec++;
		nsec-=1000L*1000L*1000L;
	}
	y->tv_sec=sec;
	y->tv_nsec=nsec;
	return(y);
}


/*! Sequence number at line buffer. */
typedef enum {
	 M6K_LINE_RANGE	=	0	/*!< 1st line. */
	,M6K_LINE_DIGIT3
	,M6K_LINE_DIGIT2
	,M6K_LINE_DIGIT1
	,M6K_LINE_DIGIT0
	,M6K_LINE_FUNCTION
	,M6K_LINE_STATUS
	,M6K_LINE_OPTION1
	,M6K_LINE_OPTION2
	,M6K_LINE_CR
	,M6K_LINE_LF
	,M6K_LINE_RANGE_2		/*!< 2nd line. */
	,M6K_LINE_DIGIT3_2
	,M6K_LINE_DIGIT2_2
	,M6K_LINE_DIGIT1_2
	,M6K_LINE_DIGIT0_2
	,M6K_LINE_FUNCTION_2
	,M6K_LINE_STATUS_2
	,M6K_LINE_OPTION1_2
	,M6K_LINE_OPTION2_2
	,M6K_LINE_CR_2
	,M6K_LINE_LF_2
	,M6K_LINE_LEN
} M6KLine;

#define	M6K_CODE_VALUE_MASK	(0x0f)

#define	M6K_LINE_BASE_MASK	(0x70)
#define	M6K_LINE_BASE		(0x30)

#define	M6K_FUNCTION_BASE	(M6K_LINE_BASE)
#define	M6K_FUNCTION_VOLTAGE	(0x0b)
#define	M6K_FUNCTION_MICRO_A	(0x0d)
#define	M6K_FUNCTION_MILLI_A	(0x0f)
#define	M6K_FUNCTION_AUTO_A	(0x00)
#define	M6K_FUNCTION_MANUAL_A	(0x09)
#define	M6K_FUNCTION_OHM	(0x03)
#define	M6K_FUNCTION_CONDUCTION	(0x05)
#define	M6K_FUNCTION_DIODE	(0x01)
#define	M6K_FUNCTION_FREQUENCY	(0x02)
#define	M6K_FUNCTION_CAPACITOR	(0x06)
#define	M6K_FUNCTION_TEMP	(0x04)
#define	M6K_FUNCTION_LUX	(0x0e)
#define	M6K_FUNCTION_DB		(0x0c)
#define	M6K_FUNCTION_ADP2	(0x08)
#define	M6K_FUNCTION_ADP3	(0x0a)
#define	M6K_FUNCTION_NUM	(0x10)
#define	M6K_FUNCTION_MASK	(0x0f)

#define	M6K_RANGE_BASE_MASK	(0x70)
#define	M6K_RANGE_BASE		(M6K_LINE_BASE)
#define	M6K_RANGE_NUM		(0x08)
#define	M6K_RANGE_MASK		(0x0f)

#define	M6K_RANGE_NONE		(127)

#define	M6K_STATUS_BASE		(M6K_LINE_BASE)
#define	M6K_STATUS_OL		(0x01)
#define	M6K_STATUS_BATT		(0x02)
#define	M6K_STATUS_SIGN		(0x04)
#define	M6K_STATUS_JUDGE	(0x08)

#define	M6K_OPTION1_BASE		(M6K_LINE_BASE)
#define	M6K_OPTION1_UNKNOWN		(0x01)
#define	M6K_OPTION1_MIN			(0x02)
#define	M6K_OPTION1_MAX			(0x04)
#define	M6K_OPTION1_HOLD		(0x08)

#define	M6K_OPTION2_BASE		(M6K_LINE_BASE)
#define	M6K_OPTION2_UNKNOWN		(0x01)
#define	M6K_OPTION2_AUTO		(0x02)
#define	M6K_OPTION2_AC			(0x04)
#define	M6K_OPTION2_DC			(0x08)

#define	M6K_LINE_CR_CODE		(0x0d)
#define	M6K_LINE_LF_CODE		(0x0a)


#define	LINE_CODE_RANGE_SET	(0x007f)	/* 0x0 to 0x6 */
#define	LINE_CODE_DIGIT_SET	(0x03ff)	/* 0x0 to 0x9 */
#define	LINE_CODE_FUNCTION_SET	((0x1<<M6K_FUNCTION_VOLTAGE)	\
				|(0x1<<M6K_FUNCTION_MICRO_A)	\
				|(0x1<<M6K_FUNCTION_MILLI_A)	\
				|(0x1<<M6K_FUNCTION_AUTO_A)	\
				|(0x1<<M6K_FUNCTION_MANUAL_A)	\
				|(0x1<<M6K_FUNCTION_OHM)	\
				|(0x1<<M6K_FUNCTION_CONDUCTION)	\
				|(0x1<<M6K_FUNCTION_DIODE)	\
				|(0x1<<M6K_FUNCTION_FREQUENCY)	\
				|(0x1<<M6K_FUNCTION_CAPACITOR)	\
				|(0x1<<M6K_FUNCTION_TEMP)	\
				|(0x1<<M6K_FUNCTION_LUX)	\
				|(0x1<<M6K_FUNCTION_DB)		\
				|(0x1<<M6K_FUNCTION_ADP2)	\
				|(0x1<<M6K_FUNCTION_ADP3)	\
				)

#define	LINE_CODE_STATUS_SET	(0xffff)
#define	LINE_CODE_OPTION1_SET	((0x1<<0x0) \
				|(0x1<<M6K_OPTION1_MIN) \
				|(0x1<<M6K_OPTION1_MAX) \
				|(0x1<<M6K_OPTION1_HOLD) \
				|(0x1<<(M6K_OPTION1_MIN+M6K_OPTION1_HOLD)) \
				|(0x1<<(M6K_OPTION1_MAX+M6K_OPTION1_HOLD)) \
				)

#define	LINE_CODE_OPTION2_SET	((0x1<<0x0) \
				|(0x1<<M6K_OPTION2_AUTO) \
				|(0x1<<M6K_OPTION2_AC) \
				|(0x1<<M6K_OPTION2_DC) \
				|(0x1<<(M6K_OPTION2_AUTO+M6K_OPTION2_AC)) \
				|(0x1<<(M6K_OPTION2_AUTO+M6K_OPTION2_DC)) \
				)

#define	LINE_CODE_CR_SET	((0x1<<M6K_LINE_CR_CODE))
#define	LINE_CODE_LF_SET	((0x1<<M6K_LINE_LF_CODE))

/*! Bitmap table to check received code. */
const uint16_t line_code_sets[M6K_LINE_LEN]={
	[M6K_LINE_RANGE]=	LINE_CODE_RANGE_SET,
	[M6K_LINE_DIGIT3]=	LINE_CODE_DIGIT_SET,
	[M6K_LINE_DIGIT2]=	LINE_CODE_DIGIT_SET,
	[M6K_LINE_DIGIT1]=	LINE_CODE_DIGIT_SET,
	[M6K_LINE_DIGIT0]=	LINE_CODE_DIGIT_SET,
	[M6K_LINE_FUNCTION]=	LINE_CODE_FUNCTION_SET,
	[M6K_LINE_STATUS]=	LINE_CODE_STATUS_SET,
	[M6K_LINE_OPTION1]=	LINE_CODE_OPTION1_SET,
	[M6K_LINE_OPTION2]	LINE_CODE_OPTION2_SET,
	[M6K_LINE_CR]=		LINE_CODE_CR_SET,
	[M6K_LINE_LF]=		LINE_CODE_LF_SET,
	[M6K_LINE_RANGE_2]=	LINE_CODE_RANGE_SET,
	[M6K_LINE_DIGIT3_2]=	LINE_CODE_DIGIT_SET,
	[M6K_LINE_DIGIT2_2]=	LINE_CODE_DIGIT_SET,
	[M6K_LINE_DIGIT1_2]=	LINE_CODE_DIGIT_SET,
	[M6K_LINE_DIGIT0_2]=	LINE_CODE_DIGIT_SET,
	[M6K_LINE_FUNCTION_2]=	LINE_CODE_FUNCTION_SET,
	[M6K_LINE_STATUS_2]=	LINE_CODE_STATUS_SET,
	[M6K_LINE_OPTION1_2]=	LINE_CODE_OPTION1_SET,
	[M6K_LINE_OPTION2_2]=	LINE_CODE_OPTION2_SET,
	[M6K_LINE_CR_2]=	LINE_CODE_CR_SET,
	[M6K_LINE_LF_2]=	LINE_CODE_LF_SET,
	};

/*! Max length of unit strings.  */
#define	TM6KPANEL_UNIT_LEN_MAX	(4)

/*! Unit table. */
const char	*function_unit[M6K_FUNCTION_NUM]={
	[M6K_FUNCTION_VOLTAGE]="V",
	[M6K_FUNCTION_MICRO_A]="A",
	[M6K_FUNCTION_MILLI_A]="A",
	[M6K_FUNCTION_AUTO_A]="A",
	[M6K_FUNCTION_MANUAL_A]="A",
	[M6K_FUNCTION_OHM]="OHM",
	[M6K_FUNCTION_CONDUCTION]="OHMc",
	[M6K_FUNCTION_DIODE]="Vf",
	[M6K_FUNCTION_FREQUENCY]="Hz",
	[M6K_FUNCTION_CAPACITOR]="F",
	[M6K_FUNCTION_TEMP]="DegC",
	[M6K_FUNCTION_LUX]="Lux",
	[M6K_FUNCTION_DB]="dB",
	[M6K_FUNCTION_ADP2]="ADP2",
	[M6K_FUNCTION_ADP3]="ADP3",
};

/*! Scientific: Voltage exponent value.
    @note four digits treat as 0.000e(this_value)
	  6.000,  60.00, 600.0, 6000,
	  600.0m, N/A,   N/A,   N/A
*/
int range_exp_volt[]={
	  0 ,1, 2, 3
	 ,-1, M6K_RANGE_NONE, M6K_RANGE_NONE, M6K_RANGE_NONE
};

/*! Scientific: Current mA exponent value.
    @note four digits treat as 0.000e(this_value)
    60.00m, 600.0m, N/A, N/A,
    N/A,    N/A,    N/A, N/A
*/
int range_exp_milli_a[]={
	 -3+1, -3+2, M6K_RANGE_NONE,M6K_RANGE_NONE
	,M6K_RANGE_NONE, M6K_RANGE_NONE, M6K_RANGE_NONE, M6K_RANGE_NONE
};

/*! Scientific: mA exponent value.
    @note four digits treat as 0.000e(this_value)
    600.0u, 6000u, N/A, N/A,
    N/A,    N/A,   N/A, N/A
*/
int range_exp_micro_a[]={
	 -6+2, -6+3, M6K_RANGE_NONE, M6K_RANGE_NONE
	,M6K_RANGE_NONE, M6K_RANGE_NONE, M6K_RANGE_NONE, M6K_RANGE_NONE
};

/*! Scientific: Auto A exponent value.
    @note four digits treat as 0.000e(this_value)
    6.000, 60.00, N/A, N/A,
    N/A,   N/A,   N/A, N/A
*/
int range_exp_auto_a[]={
	 0, 1, M6K_RANGE_NONE, M6K_RANGE_NONE
	,M6K_RANGE_NONE, M6K_RANGE_NONE, M6K_RANGE_NONE, M6K_RANGE_NONE
};

/*! Scientific: Manual A exponent value.
    @note four digits treat as 0.000e(this_value)
    60.00, N/A, N/A, N/A,
    N/A,   N/A, N/A, N/A
*/
int range_exp_manual_a[]={
	 1, M6K_RANGE_NONE, M6K_RANGE_NONE, M6K_RANGE_NONE
	,M6K_RANGE_NONE, M6K_RANGE_NONE, M6K_RANGE_NONE, M6K_RANGE_NONE
};

/*! Scientific: ohm exponent value.
    @note four digits treat as 0.000e(this_value)
    600.0,  6.000K, 60.00K, 600.0K,
    6.000M, 60.00M, N/A,    N/A
*/
int range_exp_ohm[]={
	 2, 3, 4, 5
	,6, 7, M6K_RANGE_NONE, M6K_RANGE_NONE
};

/*! Scientific: conduct ohm exponent value.
    @note four digits treat as 0.000e(this_value)
    600.0, N/A,  N/A, N/A,
    N/A,   N/A,  N/A, N/A
*/
int range_exp_conduction_ohm[]={
	 2, M6K_RANGE_NONE, M6K_RANGE_NONE, M6K_RANGE_NONE
	,M6K_RANGE_NONE, M6K_RANGE_NONE, M6K_RANGE_NONE, M6K_RANGE_NONE
};

/*! Scientific: frequency exponent value.
    @note four digits treat as 0.000e(this_value)
    6.000K, 60.00K, 600.0K, 6.000M
    60.00M, N/A,    N/A,    N/A.
*/
int range_exp_frequency[]={
	 3+0, 3+1, 3+2, 6
	,6+1, M6K_RANGE_NONE, M6K_RANGE_NONE, M6K_RANGE_NONE
};

/*! Scientific: capacitor exponent value.
    @note four digits treat as 0.000e(this_value)
    6.000n, 60.00n, 600.0n, 6.000u,
    60.00u, 600.0u, 6.000m, N/A
*/
int range_exp_capacitor[]={
	 -9, -8, -7, -6
	,-5, -4, -3, M6K_RANGE_NONE
};

/*! Scientific: diode vf exponent value.
    6.000, N/A, N/A, N/A,
    N/A,   N/A, N/A, N/A
*/
int range_exp_diode[]={
	 0, M6K_RANGE_NONE, M6K_RANGE_NONE, M6K_RANGE_NONE
	,M6K_RANGE_NONE, M6K_RANGE_NONE, M6K_RANGE_NONE, M6K_RANGE_NONE
};

/*! Scientific: Lux exponent value.
    6000, N/A, N/A, N/A,
    N/A,  N/A, N/A, N/A
*/
int range_exp_lux[]={
	 3, M6K_RANGE_NONE, M6K_RANGE_NONE, M6K_RANGE_NONE
	,M6K_RANGE_NONE, M6K_RANGE_NONE, M6K_RANGE_NONE, M6K_RANGE_NONE
};

/*! Scientific: dB exponent value.
    600.0, N/A, N/A, N/A,
    N/A,   N/A, N/A, N/A
*/
int range_exp_db[]={
	 2, M6K_RANGE_NONE, M6K_RANGE_NONE, M6K_RANGE_NONE
	,M6K_RANGE_NONE, M6K_RANGE_NONE, M6K_RANGE_NONE, M6K_RANGE_NONE
};

/*! Scientific: Temp exponent value.
    6000, N/A, N/A, N/A,
    N/A,   N/A, N/A, N/A
*/
int range_exp_temp[]={
	 3, M6K_RANGE_NONE, M6K_RANGE_NONE, M6K_RANGE_NONE
	,M6K_RANGE_NONE, M6K_RANGE_NONE, M6K_RANGE_NONE, M6K_RANGE_NONE
};

/*! Scientific exponent table at unit.
*/
int *range_exp_code[M6K_FUNCTION_NUM]={
	[M6K_FUNCTION_VOLTAGE]=range_exp_volt, 
	[M6K_FUNCTION_MICRO_A]=range_exp_micro_a,
	[M6K_FUNCTION_MILLI_A]=range_exp_milli_a,
	[M6K_FUNCTION_AUTO_A]=range_exp_auto_a,
	[M6K_FUNCTION_MANUAL_A]=range_exp_manual_a,
	[M6K_FUNCTION_OHM]=range_exp_ohm,
	[M6K_FUNCTION_CONDUCTION]=range_exp_conduction_ohm,
	[M6K_FUNCTION_DIODE]=range_exp_diode,
	[M6K_FUNCTION_FREQUENCY]=range_exp_frequency,
	[M6K_FUNCTION_CAPACITOR]=range_exp_capacitor,
	[M6K_FUNCTION_TEMP]=range_exp_temp,
	[M6K_FUNCTION_LUX]=range_exp_lux,
	[M6K_FUNCTION_DB]=range_exp_db,
};

/*! Engineering range exponent.
*/
typedef	struct {
	char	dot;	/*!< dot position. */
	char	exp;	/*!< exponent */
	char	prefix;	/*!< scaling unit character. */
} TEngFormat;

/*! Engineering: Voltage exponent value.
    6.000,  60.00, 600.0, 6000,
    600.0m, N/A,   N/A,   N/A
*/
TEngFormat range_engexp_volt[]={
	{ 1, 0, ' '}, { 2, 0, ' '}, { 3, 0, ' '}, { 4, 0, ' '},
	{ 3,-3, 'm'}, {-1, 0, ' '}, {-1, 0, ' '}, {-1, 0, ' '}
};

/*! Engineering: Current mA exponent value.
    60.00m, 600.0m, N/A, N/A,
    N/A,    N/A,    N/A, N/A
*/
TEngFormat range_engexp_milli_a[]={
	{ 2,-3, 'm'}, { 3,-3, 'm'}, {-1, 0, ' '}, {-1, 0, ' '}, 
	{-1, 0, ' '}, {-1, 0, ' '}, {-1, 0, ' '}, {-1, 0, ' '}
};

/*! Engineering: Current uA exponent value.
    600.0u, 6000u, N/A, N/A,
    N/A,    N/A,   N/A, N/A
*/
TEngFormat range_engexp_micro_a[]={
	{ 3,-6, 'u'}, { 4,-6, 'u'}, {-1, 0, ' '}, {-1, 0, ' '}, 
	{-1, 0, ' '}, {-1, 0, ' '}, {-1, 0, ' '}, {-1, 0, ' '}
};

/*! Engineering: Current Auto A exponent value.
    6.000, 60.00, N/A, N/A,
    N/A,   N/A,   N/A, N/A
*/
TEngFormat range_engexp_auto_a[]={
	{ 1, 0, ' '}, { 2, 0, ' '}, {-1, 0, ' '}, {-1, 0, ' '},
	{-1, 0, ' '}, {-1, 0, ' '}, {-1, 0, ' '}, {-1, 0, ' '}
};

/*! Engineering: Current Manual A exponent value.
    60.00, N/A, N/A, N/A
    N/A,   N/A, N/A, N/A
*/
TEngFormat range_engexp_manual_a[]={
	{ 2, 0, ' '}, {-1, 0, ' '}, {-1, 0, ' '}, {-1, 0, ' '},
	{-1, 0, ' '}, {-1, 0, ' '}, {-1, 0, ' '}, {-1, 0, ' '}
};

/*! Engineering: ohm exponent value.
    600.0,  6.000K, 60.00K, 600.0K,
    6.000M, 60.00M, N/A,    N/A
*/
TEngFormat range_engexp_ohm[]={
	{ 3, 0, ' '}, { 1, 3, 'k'}, { 2, 3, 'k'}, { 3, 3, 'k'},
	{ 1, 6, 'M'}, { 2, 6, 'M'}, {-1, 0, ' '}, {-1, 0, ' '}
};

/*! Engineering: condiction ohm exponent value.
    600.0, N/A, N/A, N/A,
    N/A,   N/A, N/A, N/A
*/
TEngFormat range_engexp_conduction_ohm[]={
	{ 3, 0, ' '}, {-1, 0, ' '}, {-1, 0, ' '}, {-1, 0, ' '},
	{-1, 0, ' '}, {-1, 0, ' '}, {-1, 0, ' '}, {-1, 0, ' '}
};

/*! Engineering: frequency exponent value.
    6.000K, 60.00K, 600.0K, 6.000M
    60.00M, N/A,    N/A,    N/A.
*/
TEngFormat range_engexp_frequency[]={
	{ 1, 3, 'k'}, { 2, 3, 'k'}, { 3, 3, 'k'}, { 1, 6, 'M'},
	{ 2, 6, 'M'}, {-1, 0, ' '}, {-1, 0, ' '}, {-1, 0, ' '}
};

/*! Engineering: capacitor exponent value.
    6.000n, 60.00n, 600.0n, 6.000u,
    60.00u, 600.0u, 6.000m, N/A
*/
TEngFormat range_engexp_capacitor[]={
	{ 1,-9, 'n'}, { 2,-9, 'n'}, { 3,-9, 'n'}, { 1,-6, 'u'},
	{ 2,-6, 'u'}, { 3,-6, 'u'}, { 1,-3, 'm'}, { 2,-3, 'm'}
};

/*! Engineering: diode vf exponent value.
    6.000, N/A, N/A, N/A,
    N/A,   N/A, N/A, N/A
*/
TEngFormat range_engexp_diode[]={
	{ 1, 0, ' '}, { 2, 0, ' '}, {-1, 0, ' '}, {-1, 0, ' '},
	{-1, 0, ' '}, {-1, 0, ' '}, {-1, 0, ' '}, {-1, 0, ' '}
};

/*! Engineering: Lux exponent value.
    6000, N/A, N/A, N/A,
    N/A,  N/A, N/A, N/A
*/
TEngFormat range_engexp_lux[]={
	{ 4, 0, ' '}, {-1, 0, ' '}, {-1, 0, ' '}, {-1, 0, ' '},
	{-1, 0, ' '}, {-1, 0, ' '}, {-1, 0, ' '}, {-1, 0, ' '}
};

/*! Engineering: dB exponent value.
    600.0, N/A, N/A, N/A,
    N/A,   N/A, N/A, N/A
*/
TEngFormat range_engexp_db[]={
	{ 3, 0, ' '}, {-1, 0, ' '}, {-1, 0, ' '}, {-1, 0, ' '},
	{-1, 0, ' '}, {-1, 0, ' '}, {-1, 0, ' '}, {-1, 0, ' '}
};

/*! Engineering: Temp exponent value.
    6000, N/A, N/A, N/A,
    N/A,   N/A, N/A, N/A
*/
TEngFormat range_engexp_temp[]={
	{ 4, 0, ' '}, {-1, 0, ' '}, {-1, 0, ' '}, {-1, 0, ' '},
	{-1, 0, ' '}, {-1, 0, ' '}, {-1, 0, ' '}, {-1, 0, ' '}
};

/*! Engineering exponent table at unit.
*/
TEngFormat *range_engexp_code[M6K_FUNCTION_NUM]={
	[M6K_FUNCTION_VOLTAGE]=range_engexp_volt, 
	[M6K_FUNCTION_MICRO_A]=range_engexp_micro_a,
	[M6K_FUNCTION_MILLI_A]=range_engexp_milli_a,
	[M6K_FUNCTION_AUTO_A]=range_engexp_auto_a,
	[M6K_FUNCTION_MANUAL_A]=range_engexp_manual_a,
	[M6K_FUNCTION_OHM]=range_engexp_ohm,
	[M6K_FUNCTION_CONDUCTION]=range_engexp_conduction_ohm,
	[M6K_FUNCTION_DIODE]=range_engexp_diode,
	[M6K_FUNCTION_FREQUENCY]=range_engexp_frequency,
	[M6K_FUNCTION_CAPACITOR]=range_engexp_capacitor,
	[M6K_FUNCTION_TEMP]=range_engexp_temp,
	[M6K_FUNCTION_LUX]=range_engexp_lux,
	[M6K_FUNCTION_DB]=range_engexp_db,
};

#define TM6K_FLAGS_CLEAR		(0x00000000)
#define TM6K_FLAGS_INVALID		(0x00000001)

#define	TM6K_PHASE_LOCK_FREE		(0)
#define	TM6K_PHASE_LOCK_LOCK		(1)
#define	TM6K_PHASE_LOCK_JUMP		(2)
#define	TM6K_PHASE_LOCK_TRACKING	(3)


/*! M-6000M Virtual Panel. */
typedef struct {
	unsigned	Flags;					/*!< Flags.				*/
	TimeSpec	TsNow;					/*!< Realtime temporal work.		*/
	TimeSpec	TsLf;					/*!< Realtime at Line Feed.		*/
	TimeSpec	Ts;					/*!< Realtime at Line starts.		*/
	TimeSpec	TsDtL1L2;				/*!< Delta time from 1st line end to 2nd line start. */
	time_t		TmNow;					/*!< Unix time temporal work.		*/
	time_t		Tm;					/*!< Unix time at Line starts.		*/
	int		Sequence;				/*!< Segment Sequence.			*/
	int		SequenceNext;				/*!< Segment Sequence next.		*/
	int		PhaseLock;				/*!< Phase lock status.			*/
	unsigned char	ReadVal[M6K_LINE_LEN];			/*!< Read value copy.			*/
	char		TsStr[TIMESPEC_SPRINT_MAX];		/*!< Realtime in string. "d10.d9"	*/
	ssize_t		TsStrLen;				/*!< String length of TsStr		*/
} TM6KPanel;

/*! Convert digit char to order number.
    @param c digit character.
    @return unsigned int <36: digit value. 0xff: Not digit character.
*/
unsigned int DigitToInt(char c)
{	if ((c>='0') && (c<='9'))
		{return(c-'0');
	}
	if ((c>='A') && (c<='Z'))
		{return(c-'A'+10);
	}
	if ((c>='a') && (c<='z')) {
		return(c-'a'+10);
	}
	return(0xff);
}

/*! Capture current time.
    @param p M-6000M Virtual panel context.
    @return int ==0: Faild calling system call, !=0: Success.
*/
int TM6KPanelCaptureNow(TM6KPanel *p)
{	int	ret;

	ret=1 /* Success. */;
	if (!TimeSpecGetRealTime(&(p->TsNow))) {
		fprintf(stderr,"%s: Can not get realtime. %s(%d).\n"
			,__func__
			,strerror(errno),errno
		);
		TimeSpecZero(&(p->TsNow));
		ret=0 /* Failed. */;
	}
	if (time(&(p->TmNow))==(time_t)-1) {
		fprintf(stderr,"%s: Can not get time. %s(%d).\n"
			,__func__
			,strerror(errno),errno
		);
		ret=0 /* Failed. */;
	}
	return ret;
}

/*! Record line start time.
    @param p M-6000M Virtual panel context.
*/
void TM6KPanelNowLineStart(TM6KPanel *p)
{	p->Ts=p->TsNow;
	p->Tm=p->TmNow;
	p->TsStrLen=TimeSpecSPrint(&(p->Ts),&(p->TsStr[0]));
	if (p->TsStrLen<0) {
		p->TsStrLen=0;
	}
}

/*! Initialize Panel
    @param p M-6000M Virtual panel context.
    @param dtl1l2 Delta time 1st line to 2nd line.
    @return int ==0: Faild calling system call, !=0: Success.
*/
int TM6KPanelInit(TM6KPanel *p, long dtl1l2)
{	int	ret;

	ret=1;
	p->Flags=TM6K_FLAGS_CLEAR;
	if (!TM6KPanelCaptureNow(p)) {
		ret=0;
	}
	if (dtl1l2<=0L) {
		/* Requested default. */
		dtl1l2=M6K_2ND_LF_MS;
	}
	if (dtl1l2>1000L) {
		/* More than 1 seconds. */
		dtl1l2=1000L;
	}
	p->TsDtL1L2.tv_sec=0;
	p->TsDtL1L2.tv_nsec=dtl1l2*1000L*1000L;
	TM6KPanelNowLineStart(p);
	p->TsLf=p->Ts;
	memset(&(p->ReadVal[0]),0,sizeof(p->ReadVal));
	p->Sequence=M6K_LINE_RANGE;
	p->SequenceNext=M6K_LINE_RANGE+1;
	p->PhaseLock=TM6K_PHASE_LOCK_FREE;
	return ret;
}

/*! Phase Lock Line Sequence.
    @param p M-6000M Virtual panel context.
    @param c read byte from serial port.
    @return int ==0: Failed, !=0: Success.
*/
int TM6KPanelPhaseLockLineSequence(TM6KPanel *p, unsigned char c)
{	TimeSpec	tsdt;
	int		ret;
	int		r;

	ret=1;
	switch (p->Sequence) {
		default:
			if (c==M6K_LINE_LF_CODE) {
				/* Receive LF code, Line end. */
				/* Jump sequence to RANGE. */
				if (CmdLine.Flags&COPT_FLAGS_DEBUG_LOCK) {
					fprintf(stderr,"%s.def: Phase lock jump. Sequence=%d\n",__func__,p->Sequence);
				}
				p->Sequence=M6K_LINE_LF_2;
				p->SequenceNext=M6K_LINE_RANGE;
				if (!TM6KPanelCaptureNow(p)) {
					p->PhaseLock=TM6K_PHASE_LOCK_FREE;
					ret=0;
					goto out;
				}
				/* Mark now, will be used to calculate dt=Trange-Tlf. */
				p->TsLf=p->TsNow;
				p->PhaseLock=TM6K_PHASE_LOCK_JUMP;
				goto out;
			}
			/* Keep phase lock state. */
			p->SequenceNext=p->Sequence+1;
			goto out;
		case M6K_LINE_RANGE:
			/* 1st line start. */
			if (!TM6KPanelCaptureNow(p)) {
				p->SequenceNext=p->Sequence+1;
				p->PhaseLock=TM6K_PHASE_LOCK_FREE;
				ret=0;
				goto out;
			}
			TM6KPanelNowLineStart(p);
			r=TimeSpecCmp(TimeSpecSub(&tsdt,&(p->TsNow),&(p->TsLf)),&(p->TsDtL1L2));
			if (r<=0) {
				/* May be 2nd line. */
				p->Sequence=M6K_LINE_RANGE_2;
				p->SequenceNext=M6K_LINE_RANGE_2+1;
				if (CmdLine.Flags&COPT_FLAGS_DEBUG_LOCK) {
					char buf[TIMESPEC_SPRINT_MAX];
					TimeSpecSPrint(&tsdt,buf);
					fprintf(stderr,"%s.range: Phase lock jump. dt=%s\n",__func__,buf);
				}
				p->PhaseLock=TM6K_PHASE_LOCK_JUMP;
				goto out;
			}
			/* Line sequence locked. */
			p->SequenceNext=p->Sequence+1;
			p->PhaseLock=TM6K_PHASE_LOCK_TRACKING;
			goto out;
		case M6K_LINE_RANGE_2:
			/* 2nd Line start. */
			if (!TM6KPanelCaptureNow(p)) {
				p->SequenceNext=p->Sequence+1;
				p->PhaseLock=TM6K_PHASE_LOCK_FREE;
				ret=0;
				goto out;
			}
			r=TimeSpecCmp(TimeSpecSub(&tsdt,&(p->TsNow),&(p->TsLf)),&(p->TsDtL1L2));
			if (r>0) {
				/* May be 1st line. */
				p->Sequence=M6K_LINE_RANGE;
				p->SequenceNext=M6K_LINE_RANGE+1;
				p->PhaseLock=TM6K_PHASE_LOCK_JUMP;
				if (CmdLine.Flags&COPT_FLAGS_DEBUG_LOCK) {
					char buf[TIMESPEC_SPRINT_MAX];
					TimeSpecSPrint(&tsdt,buf);
					fprintf(stderr,"%s.range2: Phase lock jump. dt=%s\n",__func__,buf);
				}
				goto out;
			}
			/* Line sequence locked. */
			p->SequenceNext=p->Sequence+1;
			p->PhaseLock=TM6K_PHASE_LOCK_TRACKING;
			goto out;
		case M6K_LINE_LF:
			/* 1st line end. */
			if (c==M6K_LINE_LF_CODE) {
				/* Lockon sequence. */
				if (!TM6KPanelCaptureNow(p)) {
					p->SequenceNext=p->Sequence+1;
					p->PhaseLock=TM6K_PHASE_LOCK_FREE;
					ret=0;
					goto out;
				}
				/* Mark now, will be used to calculate dt=Trange-Tlf. */
				p->TsLf=p->TsNow;
				/* Keep Phase Lock status. */
				p->SequenceNext=p->Sequence+1;
				goto out;
			}
			if (CmdLine.Flags&COPT_FLAGS_DEBUG_LOCK) {
				fprintf(stderr,"%s.lf: Phase lock free. c=%c\n",__func__,c);
			}
			p->SequenceNext=p->Sequence+1;
			p->PhaseLock=TM6K_PHASE_LOCK_FREE;
			goto out;
		case M6K_LINE_LF_2:
			/* 2nd line end. */
			if (c==M6K_LINE_LF_CODE) {
				/* Lock on sequence. */
				if (!TM6KPanelCaptureNow(p)) {
					p->SequenceNext=M6K_LINE_RANGE;
					p->PhaseLock=TM6K_PHASE_LOCK_FREE;
					ret=0;
					goto out;
				}
				/* Mark now, will be used to calculate dt=Trange-Tlf. */
				p->TsLf=p->TsNow;
				/* Keep Phase Lock status. */
				p->SequenceNext=M6K_LINE_RANGE;
				goto out;
			}
			/* Lock off sequence. */
			if (CmdLine.Flags&COPT_FLAGS_DEBUG_LOCK) {
				fprintf(stderr,"%s.lf2: Phase lock free. c=%c\n",__func__,c);
			}
			p->SequenceNext=M6K_LINE_RANGE;
			p->PhaseLock=TM6K_PHASE_LOCK_FREE;
			goto out;
	}
out:
	if (p->Sequence>=M6K_LINE_LEN) {
		p->Sequence=M6K_LINE_RANGE;
		p->SequenceNext=M6K_LINE_RANGE+1;
	}

	if (p->SequenceNext>=M6K_LINE_LEN) {
		p->SequenceNext=M6K_LINE_RANGE;
	}
	return ret;
}

/*! Store received byte into line buffer.
    @param p M-6000M Virtual panel context.
    @param c read byte from serial port.
    @return int ==0: Not complete line, !=0: Complete line.
*/
int TM6KPanelStoreLine(TM6KPanel *p, unsigned char c)
{	int		i;
	int		ret;
	unsigned	t;

	ret=0; /* Not completed. */
	i=p->Sequence;
	t=c&M6K_CODE_VALUE_MASK;
	if ((line_code_sets[i]&(0x1<<t))==0) {
		p->Flags|=TM6K_FLAGS_INVALID;
	}
	switch (i) {
		case M6K_LINE_RANGE:
			/* start receive 1st line. */
			/* Clear line buffer. */
			p->Flags=TM6K_FLAGS_CLEAR;
			memset(&(p->ReadVal[0]),0,sizeof(p->ReadVal));
			/* fall through to default. */
		default:
			if (c==M6K_LINE_CR_CODE) {
				p->Flags|=TM6K_FLAGS_INVALID;
				break;
			}
			if (c==M6K_LINE_LF_CODE) {
				p->Flags|=TM6K_FLAGS_INVALID;
				break;
			}
			if ((c&M6K_LINE_BASE_MASK)!=M6K_LINE_BASE) {
				p->Flags|=TM6K_FLAGS_INVALID;
				/* Continue. */
			}
			p->ReadVal[i]=c;
			break;
		case M6K_LINE_CR:
			/* 1st line CR. */
			if (c!=M6K_LINE_CR_CODE) {
				p->Flags|=TM6K_FLAGS_INVALID;
			}
			p->ReadVal[i]=c;
			break;
		case M6K_LINE_LF:
			/* 1st line end. */
			if (c!=M6K_LINE_LF_CODE) {
				p->Flags|=TM6K_FLAGS_INVALID;
			}
			p->ReadVal[i]=c;
			break;
		case M6K_LINE_CR_2:
			/* 2nd line CR. */
			if (c!=M6K_LINE_CR_CODE) {
				p->Flags|=TM6K_FLAGS_INVALID;
			}
			p->ReadVal[i]=c;
			break;
		case M6K_LINE_LF_2:
			/* 2nd line end. */
			if (c!=M6K_LINE_LF_CODE) {
				p->Flags|=TM6K_FLAGS_INVALID;
			}
			p->ReadVal[i]=c;
			if (memcmp(
				&(p->ReadVal[M6K_LINE_RANGE]),
				&(p->ReadVal[M6K_LINE_RANGE_2]), M6K_LINE_RANGE_2
			   )!=0) {
				/* 1st line is not match to 2nd line. */
				p->Flags|=TM6K_FLAGS_INVALID;
			}
			ret=1; /* Complete a line. */
			break;
	}
	return ret;
}


/*! Receive one byte
    @param p M-6000M Virtual panel context.
    @param c read byte from serial port.
    @return int ==0: Not complete line, !=0: Complete line.
*/
int TM6KPanelUpdateOne(TM6KPanel *p, unsigned char c)
{	int	ret;
	
	ret=0;
	TM6KPanelPhaseLockLineSequence(p,c);
	ret=TM6KPanelStoreLine(p,c);
	p->Sequence=p->SequenceNext;
	return ret;
}

/* 0123456789 */
/* +d.ddde+NN */
#define	TM6KPANEL_VALSCI_LEN_MAX	(10)

/*! %e Read Panel value in scientific.
    @param p M-6000M Virtual panel context.
    @param s points buffer to print.
    @return int <0: Failed, >=0: the number of characters printed out to s \
            which doesn't include NUL terminator.

                          0123456789
    Max s length: strlen("-9.999e+99")==10.
    s should have 8 elements or more.
*/
int TM6KPanelReadValSci(TM6KPanel *p, char *s)
{	unsigned	func;
	unsigned	range;
	int		i;
	int		*etbl;
	int		e;
	int		l;
	int		z;

	l=0;
	if ((p->ReadVal[M6K_LINE_STATUS]) & M6K_STATUS_SIGN) {
		/* with '-' sign */
		*s='-';
		s++;
		l++;
	}
	func=p->ReadVal[M6K_LINE_FUNCTION] & M6K_FUNCTION_MASK;
	range=p->ReadVal[M6K_LINE_RANGE] & M6K_RANGE_MASK;
	if (range>=M6K_RANGE_NUM) {
		/* Out of range. */
		return -1;
	}
	etbl=range_exp_code[func];
	if (etbl) {
		e=etbl[range];
		if (e>=M6K_RANGE_NONE) {
			/* Out of range. */
			return -1;
		}
	} else {
		/* Unknown range. */
		e=0;
	}
	z=0;
	while (z<(M6K_LINE_DIGIT0-M6K_LINE_DIGIT3)) {
		/* Loop while '0' continues and digit 3..1. */
		if ((p->ReadVal[z+M6K_LINE_DIGIT3])!='0') {
			break;
		}
		z++;
	}
	i=M6K_LINE_DIGIT3+z;
	*s=(char)(p->ReadVal[i]);
	s++;
	l++;
	if (i<M6K_LINE_DIGIT0) {
		*s='.';
		s++;
		l++;
		while (i<M6K_LINE_DIGIT0) {
			i++;
			*s=(char)(p->ReadVal[i]);
			s++;
		}
	}
	l+=(M6K_LINE_DIGIT0-M6K_LINE_DIGIT3)-z;
	e-=z;
	*s='e';
	s++;
	l++;
	l+=sprintf(s,"%+d",e);
	return l;
}

#define	TM6KPANEL_VALENG_LEN_MAX	(9)

/*! %E Read Panel value in engineering.
    @param p M-6000M Virtual panel context.
    @param s points buffer to print.
    @return int <0: Failed, >=0: the number of characters printed out to s \
            which doesn't include NUL terminator.

                          012345678
    Max s length: strlen("-9.999e+9")==9.
    s should have 8 elements or more.
*/
int TM6KPanelReadValEng(TM6KPanel *p, char *s)
{	unsigned	func;
	unsigned	range;
	int		i;
	TEngFormat	*etbl;
	int		e;
	int		d;
	int		l;
	int		z;

	l=0;
	if ((p->ReadVal[M6K_LINE_STATUS]) & M6K_STATUS_SIGN) {
		/* with '-' sign */
		*s='-';
		s++;
		l++;
	}
	func=p->ReadVal[M6K_LINE_FUNCTION] & M6K_FUNCTION_MASK;
	range=p->ReadVal[M6K_LINE_RANGE] & M6K_RANGE_MASK;
	if (range>=M6K_RANGE_NUM) {
		/* Out of range. */
		return -1;
	}
	etbl=range_engexp_code[func];
	if (etbl) {
		d=etbl[range].dot;
		e=etbl[range].exp;
		if (d<0) {
			/* Out of range. */
			return -1;
		}
	} else {
		/* Unknown range. */
		d=4;
		e=0;
	}
	z=0;
	while (z<(M6K_LINE_DIGIT0-M6K_LINE_DIGIT3+1)) {
		if (z==d) {
			if (z>0) {
				z--;
			}
			break;
		}
		if ((p->ReadVal[z+M6K_LINE_DIGIT3])!='0') {
			break;
		}
		z++;
	}
	if (z>=(M6K_LINE_DIGIT0-M6K_LINE_DIGIT3+1)) {
		z--;
	}
	i=z;
	while (i<=(M6K_LINE_DIGIT0-M6K_LINE_DIGIT3)) {
		if (i==d) {
			*s='.';
			s++;
			l++;
		}
		*s=(char)(p->ReadVal[i+M6K_LINE_DIGIT3]);
		s++;
		i++;
	}
	*s='e';
	s++;
	if (e<0) {
		*s='-';
		s++;
		*s='0'-e;
		s++;
	} else {
		*s='+';
		s++;
		*s='0'+e;
		s++;
	}
	l+=(M6K_LINE_DIGIT0-M6K_LINE_DIGIT3+1)+3-z;
	*s=0;
	return l;
}

#define	TM6KPANEL_VALVIS_LEN_MAX	(6)

/*! %v Read Panel value in visual style.
    @param p M-6000M Virtual panel context.
    @param s points buffer to print.
    @return int <0: Failed, >=0: the number of characters printed out to s \
            which doesn't include NUL terminator.

                          012345
    Max s length: strlen("-9.999")==6.
    s should have 8 elements or more.
*/
int TM6KPanelReadValVis(TM6KPanel *p, char *s)
{	unsigned	func;
	unsigned	range;
	int		i;
	TEngFormat	*etbl;
	int		d;
	int		l;
	int		z;

	l=0;
	if ((p->ReadVal[M6K_LINE_STATUS]) & M6K_STATUS_SIGN) {
		/* with '-' sign */
		*s='-';
		s++;
		l++;
	}
	func=p->ReadVal[M6K_LINE_FUNCTION] & M6K_FUNCTION_MASK;
	range=p->ReadVal[M6K_LINE_RANGE] & M6K_RANGE_MASK;
	if (range>=M6K_RANGE_NUM) {
		/* Out of range. */
		return -1;
	}
	etbl=range_engexp_code[func];
	if (etbl) {
		d=etbl[range].dot;
		if (d<0) {
			/* Out of range. */
			return 0;
		}
	} else {
		/* Unknown range. */
		d=4;
	}
	z=0;
	while (z<(M6K_LINE_DIGIT0-M6K_LINE_DIGIT3+1)) {
		if (z==d) {
			if (z>0) {
				z--;
			}
			break;
		}
		if ((p->ReadVal[z+M6K_LINE_DIGIT3])!='0') {
			break;
		}
		z++;
	}
	if (z>=(M6K_LINE_DIGIT0-M6K_LINE_DIGIT3+1)) {
		z--;
	}
	i=z;
	while (i<=(M6K_LINE_DIGIT0-M6K_LINE_DIGIT3)) {
		if (i==d) {
			*s='.';
			s++;
			l++;
		}
		*s=(char)(p->ReadVal[i+M6K_LINE_DIGIT3]);
		s++;
		i++;
	}
	l+=(M6K_LINE_DIGIT0-M6K_LINE_DIGIT3+1-z);
	*s=0;
	return l;
}

#define	TM6KPANEL_ACDC_LEN_MAX	(2)

/*! %a Read AC/DC mode.
    @param p M-6000M Virtual panel context.
    @param s points buffer to print.
    @param bs blank character.
    @return int <0: Failed, >=0: the number of characters printed out to s \
            which doesn't include NUL terminator.
*/
int TM6KPanelReadAcDc(TM6KPanel *p, char *s, char bs)
{	unsigned char	o2;

	o2=p->ReadVal[M6K_LINE_OPTION2];

	if (o2&M6K_OPTION2_AC) {
		*s='A';
		*(s+1)='C';
		*(s+2)=0;
		return 2;
	}

	if (o2&M6K_OPTION2_DC) {
		*s='D';
		*(s+1)='C';
		*(s+2)=0;
		return 2;
	}
	/* Not AC, Not DC. */
	if (bs==0) {
		*s=bs;
		return 0;
	}
	*s=bs;
	s++;
	*s=bs;
	s++;
	*s=0;
	return 2;
}

/*! %u Read Panel Unit.
    @param p M-6000M Virtual panel context.
    @param s points buffer to print.
    @param bs blank character.
    @return int <0: Failed, >=0: the number of characters printed out to s \
            which doesn't include NUL terminator.
*/
int TM6KPanelReadUnit(TM6KPanel *p, char *s, char bs)
{	const char	_sTc[]="DegC";
	const char	_sTf[]="DegF";
	int		func;
	const char	*u;
	int		l;

	func=p->ReadVal[M6K_LINE_FUNCTION]&M6K_FUNCTION_MASK;
	if (func==M6K_FUNCTION_TEMP) {
		if ((p->ReadVal[M6K_LINE_STATUS]&M6K_STATUS_JUDGE)!=0) {
			memcpy(s,_sTc,sizeof(_sTc));
			return sizeof(_sTc)-1;
		} else {
			memcpy(s,_sTf,sizeof(_sTf));
			return sizeof(_sTf)-1;
		}
	}
	u=function_unit[func];
	if (!u) {
		if (bs==0) {
			*s=bs;
			return 0;
		}
		*s=bs;
		s++;
		return 1;
	}
	l=strlen(u);
	memcpy(s,u,l+1);
	return l;
}

#define	TM6KPANEL_PREFIX_LEN_MAX	(1)

/*! %p Read Panel Prefix.
    @param p M-6000M Virtual panel context.
    @param s points buffer to print.
    @return int <0: Failed, >=0: the number of characters printed out to s \
            which doesn't include NUL terminator.
*/
int TM6KPanelReadPrefix(TM6KPanel *p, char *s)
{	unsigned	func;
	unsigned	range;
	int		l;
	TEngFormat	*etbl;
	int		d;
	char		prefix;

	l=0;
	func=p->ReadVal[M6K_LINE_FUNCTION] & M6K_FUNCTION_MASK;
	range=p->ReadVal[M6K_LINE_RANGE] & M6K_RANGE_MASK;
	if (range >= M6K_RANGE_NUM) {
		/* Out of range. */
		fprintf(stderr, "Prefix.1: Out of range. range=0x%.2x, func=0x%.2x\n"
			,p->ReadVal[M6K_LINE_RANGE]
			,p->ReadVal[M6K_LINE_FUNCTION]
		);
		return -1;
	}
	etbl=range_engexp_code[func];
	if (etbl) {
		d=etbl[range].dot;
		prefix=etbl[range].prefix;
		if (d<0) {
			/* Out of range. */
			fprintf(stderr, "Prefix.2: Out of range. range=0x%.2x, func=0x%.2x\n"
				,p->ReadVal[M6K_LINE_RANGE]
				,p->ReadVal[M6K_LINE_FUNCTION]
			);
			return -1;
		}
	} else {
		/* Unknown range. */
		return 0;
	}
	if (prefix!=' ') {
		*s=prefix;
		s++;
		l++;
	}
	*s=0;
	return l;
}

#define	TM6KPANEL_HOLD_LEN_MAX	(4)

/*! %h Read Panel Mode Hold.
    @param p M-6000M Virtual panel context.
    @param s points buffer to print.
    @return int <0: Failed, >=0: the number of characters printed out to s \
            which doesn't include NUL terminator.
*/
int TM6KPanelReadHold(TM6KPanel *p, char *s, char bs)
{	const char _sHold[]="HOLD";

	if (p->ReadVal[M6K_LINE_OPTION1] & M6K_OPTION1_HOLD) {
		memcpy(s,_sHold,sizeof(_sHold));
		return(sizeof(_sHold)-1);
	}
	if (bs==0) {
		*s=bs;
		return 0;
	}
	memset(s,bs,sizeof(_sHold)-1);
	*(s+sizeof(_sHold)-1)=0;
	return sizeof(_sHold)-1;
}

#define	TM6KPANEL_MINMAX_LEN_MAX	(3)

/*! %m Read Panel Mode Min Max.
    @param p M-6000M Virtual panel context.
    @param s points buffer to print.
    @param bs blank character.
    @return int <0: Failed, >=0: the number of characters printed out to s \
            which doesn't include NUL terminator.
*/
int TM6KPanelReadMinMax(TM6KPanel *p, char *s, char bs)
{	const char _sMin[]="MIN";
	const char _sMax[]="MAX";
	unsigned char	o1;

	o1=p->ReadVal[M6K_LINE_OPTION1];
	if (o1&M6K_OPTION1_MIN) {
		memcpy(s,_sMin,sizeof(_sMin));
		return(sizeof(_sMin)-1);
	}
	if (o1&M6K_OPTION1_MAX) {
		memcpy(s,_sMax,sizeof(_sMax));
		return(sizeof(_sMax)-1);
	}
	if (bs==0) {
		*s=bs;
		return 0;
	}
	memset(s,bs,sizeof(_sMin)-1);
	*(s+sizeof(_sMin)-1)=0;
	return sizeof(_sMin)-1;
}

#define	TM6KPANEL_AUTO_LEN_MAX	(4)

/*! %n Read Panel Mode Auto.
    @param p M-6000M Virtual panel context.
    @param s points buffer to print.
    @param bs blank character.
    @return int <0: Failed, >=0: the number of characters printed out to s \
            which doesn't include NUL terminator.
*/
int TM6KPanelReadAuto(TM6KPanel *p, char *s, char bs)
{	const char _sAuto[]="AUTO";
	unsigned char	o2;

	o2=p->ReadVal[M6K_LINE_OPTION2];
	if (o2&M6K_OPTION2_AUTO) {
		memcpy(s,_sAuto,sizeof(_sAuto));
		return(sizeof(_sAuto)-1);
	}
	if (bs==0) {
		*s=bs;
		return 0;
	}
	memset(s,bs,sizeof(_sAuto)-1);
	*(s+sizeof(_sAuto)-1)=0;
	return sizeof(_sAuto)-1;
}

#define	TM6KPANEL_JUDGE_LEN_MAX	(5)

/*! %j Read Panel Judge
    @param p M-6000M Virtual panel context.
    @param s points buffer to print.
    @param bs blank character.
    @return int <0: Failed, >=0: the number of characters printed out to s \
            which doesn't include NUL terminator.
*/
int TM6KPanelReadJudge(TM6KPanel *p, char *s, char bs)
{	const char _sJudge[]="JUDGE";
	if (p->ReadVal[M6K_LINE_STATUS]&M6K_STATUS_JUDGE) {
		memcpy(s,_sJudge,sizeof(_sJudge));
		return(sizeof(_sJudge)-1);
	}
	if (bs==0) {
		*s=bs;
		return 0;
	}
	memset(s,bs,sizeof(_sJudge)-1);
	*(s+sizeof(_sJudge)-1)=0;
	return sizeof(_sJudge)-1;
}

#define	TM6KPANEL_BATT_LEN_MAX	(4)

/*! %b Read Panel Batt.
    @param p M-6000M Virtual panel context.
    @param s points buffer to print.
    @param bs blank character.
    @return int <0: Failed, >=0: the number of characters printed out to s \
            which doesn't include NUL terminator.
*/
int TM6KPanelReadBatt(TM6KPanel *p, char *s, char bs)
{	const char _sBatt[]="BATT";
	if (p->ReadVal[M6K_LINE_STATUS]&M6K_STATUS_BATT) {
		memcpy(s,_sBatt,sizeof(_sBatt));
		return(sizeof(_sBatt)-1);
	}
	if (bs==0) {
		*s=bs;
		return 0;
	}
	memset(s,bs,sizeof(_sBatt)-1);
	*(s+sizeof(_sBatt)-1)=0;
	return sizeof(_sBatt)-1;
}

#define	TM6KPANEL_OL_LEN_MAX	(2)

/*! %o Read Panel OL.
    @param p M-6000M Virtual panel context.
    @param s points buffer to print.
    @param bs blank character.
    @return int <0: Failed, >=0: the number of characters printed out to s \
            which doesn't include NUL terminator.
*/
int TM6KPanelReadOL(TM6KPanel *p, char *s, char bs)
{	const char _sOl[]="OL";
	if (p->ReadVal[M6K_LINE_STATUS]&M6K_STATUS_OL) {
		memcpy(s,_sOl,sizeof(_sOl));
		return(sizeof(_sOl)-1);
	}
	if (bs==0) {
		*s=bs;
		return 0;
	}
	memset(s,bs,sizeof(_sOl)-1);
	*(s+sizeof(_sOl)-1)=0;
	return sizeof(_sOl)-1;
}

#define	TM6KPANEL_RAW_LEN_MAX	(M6K_LINE_CR)

/*! %Z Read raw in string.
    @param p M-6000M Virtual panel context.
    @param s points buffer to print.
    @return int <0: Failed, >=0: the number of characters printed out to s \
            which doesn't include NUL terminator.
*/
int TM6KPanelReadRaw(TM6KPanel *p, char *s)
{	memcpy(s,&(p->ReadVal[0]), M6K_LINE_CR);
	*(s+M6K_LINE_CR)=0;
	return TM6KPANEL_RAW_LEN_MAX;
}

#define	TM6KPANEL_RAWHEX_LEN_MAX	(M6K_LINE_LEN*3-1)

/*! %X Read raw in hex.
    @param p M-6000M Virtual panel context.
    @param s points buffer to print.
    @return int <0: Failed, >=0: the number of characters printed out to s \
            which doesn't include NUL terminator.
*/
int TM6KPanelReadRawHex(TM6KPanel *p, char *s)
{			     /*r3210fsopCL*/
	const char _hex_sep[]=":...:::.:.";
	int	i;

	const char	*sep;
	unsigned char	*rv;

	rv=&(p->ReadVal[0]);
	sep=_hex_sep;
	i=0;
	while (i<M6K_LINE_RANGE_2) {
		char	c;
		c=*sep;
		if (c) {
			s+=sprintf(s,"%.2x%c",*rv,c);
		} else {
			s+=sprintf(s,"%.2x",*rv);
		}
		rv++;
		sep++;
		i++;
	}
	*s=0;
	return TM6KPANEL_RAWHEX_LEN_MAX;
}

/* sprintf M-6000M Reads.
   @param p M-6000M Virtual panel context.
   @param fmt format string.
   @param s buffer to print.
   @param len0 buffer length with NUL terminator pointed by s.
   @return int printed length with NUL terminator.
   sprintf format
   %a AC/DC
   %v reading value without prefix.
   %E reading value with engineering-exponential (e-9, e-6, .. ,e+6).
   %e reading value with scientific-exponential.
   %p prefix.
   %u unit.
   %h hold.
   %m min/max.
   %j judge.
   %o over.
   %n auto-range.
   %b battery.
   %T time reads in time_t
   %R real time reads in nano seconds.
   %X raw read in hex.
   %Z raw read in string.
   @return int formatted size (including NUL terminator).
*/
int TM6KPanelSPrintf(TM6KPanel *p, char *fmt, char *s, int len0)
{	int	part;
	int	len;
	char	c;

	len=len0;
	if (len>0) {
		/* buffer s has one or more elements. */
		/* reserve NUL terminate char. */
		len--;
	} else	{/* buffer s has no element. */
		return(0);
	}
	while (len>0) {
		part=0;
		c=*fmt;
		if (c==0) {
			/* End of format. */
			break;
		}
		if (c=='%') {
			/* format escape. */
			fmt++;
			c=*fmt;
			switch (c) {
		 		case 'a': {
					/* AC/DC */
					if (len<TM6KPANEL_ACDC_LEN_MAX) {
						break;
					}
					part=TM6KPanelReadAcDc(p,s,CmdLine.Blank);
					if (part<0) {
						p->Flags|=TM6K_FLAGS_INVALID;
						fmt++;
						break;
					}
					s+=part;
					len-=part;
					fmt++;
					break;
				}
				case 'v': {
					/* Value in visual style. */
					if (len<TM6KPANEL_VALVIS_LEN_MAX) {
						break;
					}
					part=TM6KPanelReadValVis(p,s);
					if (part<0) {
						p->Flags|=TM6K_FLAGS_INVALID;
						fmt++;
						break;
					}
					s+=part;
					len-=part;
					fmt++;
					break;
				}
				case 'e': {
					/* Value with scientific exponential.  */
					if (len<TM6KPANEL_VALSCI_LEN_MAX) {
						break;
					}
					part=TM6KPanelReadValSci(p,s);
					if (part<0) {
						p->Flags|=TM6K_FLAGS_INVALID;
						fmt++;
						break;
					}
					s+=part;
					len-=part;
					fmt++;
					break;
				}
				case 'E': {
					/* Value with engineering exponential.  */
					if (len<TM6KPANEL_VALENG_LEN_MAX) {
						break;
					}
					part=TM6KPanelReadValEng(p,s);
					if (part<0) {
						p->Flags|=TM6K_FLAGS_INVALID;
						fmt++;
						break;
					}
					s+=part;
					len-=part;
					fmt++;
					break;
				}
				case 'p': {
					/* Prefix, one of u, m, K, M. */
					if (len<TM6KPANEL_PREFIX_LEN_MAX) {
						break;
					}
					part=TM6KPanelReadPrefix(p,s);
					if (part<0) {
						p->Flags|=TM6K_FLAGS_INVALID;
						fmt++;
						break;
					}
					s+=part;
					len-=part;
					fmt++;
					break;
				}
				case 'u': {
					/* Unit. */
					if (len<TM6KPANEL_UNIT_LEN_MAX) {
						break;
					}
					part=TM6KPanelReadUnit(p,s,CmdLine.Blank);
					if (part<0) {
						p->Flags|=TM6K_FLAGS_INVALID;
						fmt++;
						break;
					}
					s+=part;
					len-=part;
					fmt++;
					break;
				}
				case 'h': {
					/* Hold. */
					if (len<TM6KPANEL_HOLD_LEN_MAX) {
						break;
					}
					part=TM6KPanelReadHold(p,s,CmdLine.Blank);
					if (part<0) {
						p->Flags|=TM6K_FLAGS_INVALID;
						fmt++;
						break;
					}
					s+=part;
					len-=part;
					fmt++;
					break;
				}
				case 'm': {
					/* min/max */
					if (len<TM6KPANEL_MINMAX_LEN_MAX) {
						break;
					}
					part=TM6KPanelReadMinMax(p,s,CmdLine.Blank);
					if (part<0) {
						p->Flags|=TM6K_FLAGS_INVALID;
						fmt++;
						break;
					}
					s+=part;
					len-=part;
					fmt++;
					break;
				}
				case 'j': {
					/* Judge. */
					if (len<TM6KPANEL_JUDGE_LEN_MAX) {
						break;
					}
					part=TM6KPanelReadJudge(p,s,CmdLine.Blank);
					if (part<0) {
						p->Flags|=TM6K_FLAGS_INVALID;
						fmt++;
						break;
					}
					s+=part;
					len-=part;
					fmt++;
					break;
				}
				case 'o': {
					/* Over. OL */
					if (len<TM6KPANEL_OL_LEN_MAX) {
						break;
					}
					part=TM6KPanelReadOL(p,s,CmdLine.Blank);
					if (part<0) {
						p->Flags|=TM6K_FLAGS_INVALID;
						fmt++;
						break;
					}
					s+=part;
					len-=part;
					fmt++;
					break;
				}
				case 'n': {
					/* auto-range. */
					if (len<TM6KPANEL_AUTO_LEN_MAX) {
						break;
					}
					part=TM6KPanelReadAuto(p,s,CmdLine.Blank);
					if (part<0) {
						p->Flags|=TM6K_FLAGS_INVALID;
						fmt++;
						break;
					}
					s+=part;
					len-=part;
					fmt++;
					break;
				}
				case 'b': {
					/* battery. */
					if (len<TM6KPANEL_BATT_LEN_MAX) {
						break;
					}
					part=TM6KPanelReadBatt(p,s,CmdLine.Blank);
					if (part<0) {
						p->Flags|=TM6K_FLAGS_INVALID;
						fmt++;
						break;
					}
					s+=part;
					len-=part;
					fmt++;
					break;
				}
				 case 'T': {
					/* Time */
					int l;
				
					if (len<3*sizeof(time_t)) {
						break;
					}
					l=sprintf(s,"%llu",(unsigned long long)(p->Tm));
					s+=l;
					len-=l;
					fmt++;
					break;
				}
				case 'R': {
					/* Real Time. */
				
					if (len<TIMESPEC_SPRINT_MAX) {
						break;
					}
					memcpy(s, &(p->TsStr), p->TsStrLen);
					s+=p->TsStrLen;
					len-=p->TsStrLen;
					fmt++;
					break;
				}
				case 'X': {
					/* Raw reads in hex */
					if (len<TM6KPANEL_RAWHEX_LEN_MAX) {
						break;
					}
					part=TM6KPanelReadRawHex(p,s);
					if (part<0) {
						p->Flags|=TM6K_FLAGS_INVALID;
						fmt++;
						break;
					}
					s+=part;
					len-=part;
					fmt++;
					break;
				}
				case 'Z': {
					/* Raw reads. */
					if (len<TM6KPANEL_RAW_LEN_MAX) {
						break;
					}
					part=TM6KPanelReadRaw(p,s);
					if (part<0) {
						p->Flags|=TM6K_FLAGS_INVALID;
						fmt++;
						break;
					}
					s+=part;
					len-=part;
					fmt++;
					break;
				}
				case '\0': {
					break;
				}
				case '%': {
					*s=c;
					s++;
					len--;
					fmt++;
					break;
				}
				default: {
					*s=c;
					s++;
					len--;
					fmt++;
					break;
				}
			}
		} else {
			/* copy */
			part=1;
			*s=c;
			s++;
			len--;
			fmt++;
		}
	}
	*s=0;
	return(len0-len);
}

/* Estimate formatted output buffer size.
   @param p M-6000M virtual panel context.
   @param fmt M-6000M panel output format.
   @return int formatted output string max length with NUL terminator.

   sprintf format
   %a AC/DC
   %v reading value without prefix.
   %E reading value with engineering-exponential (e-9, e-6, .. ,e+6).
   %e reading value with scientific-exponential.
   %p prefix.
   %u unit.
   %h hold.
   %m min/max.
   %j judge.
   %o over.
   %n auto.
   %b battery.
   %T time reads in time_t
   %R real time reads in nano seconds.
   %X raw read in hex.
   %Z raw read in string.

*/
int TM6KPanelSPrintfSize(TM6KPanel *p, char *fmt)
{	int	len;
	char	c;

	len=0;
	while (*fmt) {
		c=*fmt;
		if (c==0) {
			/* End of format. */
			break;
		}
		if (c=='%') {
			/* format escape. */
			fmt++;
			c=*fmt;
			switch (c) {
				case 'a':
					len+=TM6KPANEL_ACDC_LEN_MAX;
					fmt++;
					break;
				case 'v':
					len+=TM6KPANEL_VALVIS_LEN_MAX;
					fmt++;
					break;
				case 'e':
					len+=TM6KPANEL_VALSCI_LEN_MAX;
					fmt++;
					break;
				case 'E':
					len+=TM6KPANEL_VALENG_LEN_MAX;
					fmt++;
					break;
				case 'p':
					len+=TM6KPANEL_PREFIX_LEN_MAX;
					fmt++;
					break;
				case 'u':
					len+=TM6KPANEL_UNIT_LEN_MAX;
					fmt++;
					break;
				case 'h':
					len+=TM6KPANEL_HOLD_LEN_MAX;
					fmt++;
					break;
				case 'm':
					len+=TM6KPANEL_MINMAX_LEN_MAX;
					fmt++;
					break;
				case 'j':
					len+=TM6KPANEL_JUDGE_LEN_MAX;
					fmt++;
					break;
				case 'o':
					len+=TM6KPANEL_OL_LEN_MAX;
					fmt++;
					break;
				case 'n':
					len+=TM6KPANEL_AUTO_LEN_MAX;
					fmt++;
					break;
				case 'b':
					len+=TM6KPANEL_BATT_LEN_MAX;
					fmt++;
					break;
				case 'T':
					len+=3*sizeof(time_t);
					fmt++;
					break;
				case 'R':
					len+=TIMESPEC_SPRINT_MAX;
					fmt++;
					break;
				case 'Z':
					len+=TM6KPANEL_RAW_LEN_MAX;
					fmt++;
					break;
				case 'X':
					len+=TM6KPANEL_RAWHEX_LEN_MAX;
					fmt++;
					break;
				case '\0':
					break;
				case '%':
					len++;
					fmt++;
					break;
				default:
					len++;
					fmt++;
					break;
			}
		} else {
			len++;
			fmt++;
		}
	}
	len++; /* NUL terminator. */
	return(len);
}

/*! File descriptor number means invalid. */
#define	INVALID_FD	(-1)

/*! serial port context, path and configuration.
    @note SicPtr[]={&(Sic[0]),&(Sic[1]),&(Sic[0])};
*/
typedef struct {
	char			*Path;			/*!< port path.	*/
	int			Fd;			/*!< port file descriptor. */
	struct termios		TermiosSave;		/*!< saved termio configuration. */
	struct termios		TermiosCur;		/*!< current termio configuration. */
	int			SicIndex;		/*!< Serial icounter. [0..1] */
	struct serial_icounter_struct	*SicPtr[3];	/*!< Serial icounter pointer array */
	struct serial_icounter_struct	Sic[2];		/*!< Serial icounter */
	char			ReadByte;		/*!< Read byte. */
} TSerialPort;

/*! Init serial port context.
    @param p serial port context.
    @return int !=0: Success, ==0: Failed.
*/
int TSerialPortInit(TSerialPort *p)
{	memset(p,0,sizeof(*p));
	p->Fd=INVALID_FD;
	p->SicIndex=0;
	p->SicPtr[0]=&(p->Sic[0]);
	p->SicPtr[1]=&(p->Sic[1]);
	p->SicPtr[2]=&(p->Sic[0]);
	return(1);
}

/*! Close serial port.
    @param p serial port context.
    @return int !=0: Success, ==0: Failed.
*/
int TSerialPortClose(TSerialPort *p)
{	int	ret;

	ret=1 /* success. */;
	if (p->Fd!=INVALID_FD) {
		/* opened. */
		/* Restore Terminal configuration. */
		if (tcsetattr(p->Fd,TCSANOW,&(p->TermiosSave))!=0) {
			fprintf(stderr,"%s: Can not restore configuration. %s(%d).\n"
				,p->Path
				,strerror(errno),errno
			);
			ret=0 /* failed. */;
		}
		if (close(p->Fd)!=0) {
			fprintf(stderr,"%s: Can not close. %s(%d).\n"
				,p->Path
				,strerror(errno),errno
				);
			ret=0 /* failed. */;
		}
		p->Fd=INVALID_FD;
	}
	if (p->Path) {
		/* Allocated path. */
		free(p->Path);
		p->Path=NULL;
	}
	return ret;
}

/*! Capture error counters.
    @param p serial port context.
    @return int !=0: Success, ==0: Failed.
*/
int TSerialPortCaptureCounters(TSerialPort *p)
{	int	ret;
	int	r;
	struct serial_icounter_struct	*sic;

	ret=1;
	sic=p->SicPtr[p->SicIndex];
	r=ioctl(p->Fd,TIOCGICOUNT,(uintptr_t)(sic));
	if (r!=0) {
		fprintf(stderr,"%s: Can not capture error counters. %s(%d).\n"
			,p->Path
			,strerror(errno),errno
		);
		ret=0;
	}
	return ret;
}

/*! open serial port.
    @param p serial port context.
    @param path serial port path.
    @return int !=0: Success, ==0: Failed.
*/
int TSerialPortOpen(TSerialPort *p, char *Path)
{	int	fd;

	if (p->Fd!=INVALID_FD) {
		TSerialPortClose(p);
	}
	p->Path=strdup(Path);
	if (!(p->Path)) {
		fprintf(stderr,"%s: Can not allocate memory. %s(%d).\n",__func__
			,strerror(errno),errno
		);
		return(0);
	}
	fd=open(Path,O_RDWR,0);
	if (fd<0) {
		fprintf(stderr,"%s: Can not open. %s(%d).\n"
			,p->Path,strerror(errno),errno
		);
		goto out_err_open;
	}
	p->Fd=fd;
	if (tcgetattr(p->Fd,&(p->TermiosSave))!=0) {
		fprintf(stderr,"%s: Can not get terminal configuration. %s(%d).\n"
			,p->Path,strerror(errno),errno
		);
		goto out_err_getattr;
	}
	memcpy(&(p->TermiosCur),&(p->TermiosSave),sizeof(p->TermiosCur));
	return(1);
out_err_getattr:
	close(p->Fd);
out_err_open:
	free(p->Path);
	p->Path=NULL;
	return(0);
}

/*! Configure serial port.
    @param p serial port context.
    @return int !=0: Success, ==0: Failed.
*/
int TSerialPortConfigM6K(TSerialPort *p)
{	int	ret;

	ret=1;
	cfmakeraw(&(p->TermiosCur));
	if (cfsetispeed(&(p->TermiosCur),B19200)!=0) {
		fprintf(stderr,"%s: Can not set input speed. %s(%d).\n"
			,p->Path,strerror(errno),errno
		);
		return(0);
	}
	if (cfsetospeed(&(p->TermiosCur),B19200)!=0) {
		fprintf(stderr,"%s: Can not set output speed. %s(%d).\n"
			,p->Path,strerror(errno),errno
		);
		return(0);
	}
	p->TermiosCur.c_iflag&=~
		(IGNPAR
		|PARMRK
		|IXON
		|IXOFF
		|IMAXBEL
		|IUTF8
		);
	p->TermiosCur.c_cflag&=~
		(CSIZE
		|CSTOPB
		|CRTSCTS
		);
	/* 7bits, odd parity */
	p->TermiosCur.c_cflag|=
		(CS7
		|CREAD
		|PARENB
		|PARODD
		|CLOCAL
		);
	p->TermiosCur.c_lflag&=~
		(ISIG
		|ICANON
		|XCASE
		|ECHO
		|ECHOE
		|ECHOK
		|ECHONL
		|ECHOCTL
		|ECHOPRT
		|ECHOKE
		|FLUSHO
		|NOFLSH
		|TOSTOP
		|PENDIN
		|IEXTEN
		);

	if (tcsetattr(p->Fd,TCSANOW,&(p->TermiosCur))!=0) {
		fprintf(stderr,"%s: Can not set configuration. %s(%d).\n"
			,p->Path
			,strerror(errno),errno
		);
		ret=0;
	}
	return(ret);
}

/*! Read bytes from serial port.
    @param p serial port context.
    @param buf buffer to store read bytes.
    @param len buffer length pointed by buf.
    @return int !=0: Success, ==0: Failed.
*/
int TSerialPortReadBytes(TSerialPort *p, char *buf, int len)
{	int	r;
	int	ret;

	ret=1;
	r=read(p->Fd,buf,len);
	if (r!=0) {
		fprintf(stderr,"%s: Can not read. %s(%d).\n"
			,p->Path
			,strerror(errno),errno
		);
		ret=0;
	}
	p->SicIndex^=0x1;
	p->SicIndex&=0x1;
	if (!TSerialPortCaptureCounters(p)) {
		ret=0;
	}
	return ret;
}

/*! Detect serial port error.
    @param p serial port context.
    @return int !=0: Detected one or more error, ==0: No error.
*/
int TSerialPortDetectCountersError(TSerialPort *p)
{	struct serial_icounter_struct	*sic_prev;
	struct serial_icounter_struct	*sic_cur;
	int				err;
	
	err=0;
	sic_cur=p->SicPtr[p->SicIndex];
	sic_prev=p->SicPtr[p->SicIndex+1];
	if (sic_cur->frame!=sic_prev->frame) {
		fprintf(stderr, "%s: Framing error. frame=%d\n",p->Path, sic_cur->frame);
		err=1;
	}
	if (sic_cur->overrun!=sic_prev->overrun) {
		fprintf(stderr, "%s: Overrun error. overrun=%d\n",p->Path, sic_cur->overrun);
		err=1;
	}
	if (sic_cur->parity!=sic_prev->parity) {
		fprintf(stderr, "%s: Parity error. parity=%d\n",p->Path, sic_cur->parity);
		err=1;
	}
	if (sic_cur->brk!=sic_prev->brk) {
		fprintf(stderr, "%s: Break. brk=%d\n",p->Path, sic_cur->brk);
		err=1;
	}
	if (sic_cur->buf_overrun!=sic_prev->buf_overrun) {
		fprintf(stderr, "%s: Buffer overrun. buf_overrun=%d\n",p->Path, sic_cur->buf_overrun);
		err=1;
	}
	return err;
}

/*! Send break.
    @param p serial port context.
    @return int !=0: Success, ==0: Failed.
*/
int TSerialPortSendBreak(TSerialPort *p)
{	int	r;
	int	ret;

	ret=1;
	r=tcsendbreak(p->Fd, 0);
	if (r!=0) {
		fprintf(stderr,"%s: Can not send break. %s(%d).\n"
			,p->Path
			,strerror(errno),errno
		);
		ret=0;
	}
	return ret;
}


#define	WRITE_SLOWDOWN_SLEEP	(10*(useconds_t)1000)

/*! Write to file descriptor with retry.
    @param fd file descriptor to write.
    @param buf buffer to write.
    @param len byte length to write.
    @return ssize_t length of written bytes.
    
*/
ssize_t WriteRetry(int fd, void *buf, ssize_t len)
{	ssize_t	remain;
	ssize_t	written_len;
	ssize_t	retry;
	ssize_t	slow_down;

	remain=len;
	retry=len;
	slow_down=len>>2;

	while ((remain>0) && (retry>0)) {
		written_len=write(fd,buf,remain);
		if (written_len<0) { 
			fprintf(stderr,"%s: Can not write. %s(%d).\n",__func__,strerror(errno),errno);
			return written_len;
		}
		if (written_len==0) {
			fprintf(stderr,"%s: Output device full? %s(%d).\n",__func__,strerror(errno),errno);
			return written_len;
		}
		buf+=written_len;
		remain-=written_len;
		if (remain<=0) {
			break;
		}
		if (slow_down>0) {
			slow_down--;
		}
		  else	{
			usleep(WRITE_SLOWDOWN_SLEEP);
		}
		 retry--;
	}
	return(len-remain);
}

#define	MAINB_CLEAR_QUEUE_FINISHED	(INT_MAX)
#define	MAINB_CLEAR_QUEUE_START		(0)

#define	MAINB_CLEAR_QUEUE_TIMEOUT	(100)

/*! main function.
    @return int ==0: Failed, !=0: Success.
*/
int MainB(void)
{	int		ret;
	int		poll_ret;

	struct pollfd	pfds[2];
	int		buf_len;
	char		*buf;
	char		buf_stdin[LINE_MAX+sizeof(long)];
	TM6KPanel	m6k;
	TSerialPort	port;
	int		clear_queue;
	long		break_retry;

	ret=1;
	buf=NULL;
	memset(&m6k, 0, sizeof(m6k));
	TSerialPortInit(&port);
	if (!TSerialPortOpen(&port,CmdLine.PortPath)) {
		ret=0;
		goto out;
	}

	if (!TSerialPortConfigM6K(&port)) {
		ret=0;
		goto out_close;
	}

	if (CmdLine.Flags&COPT_FLAGS_BREAK_AT_START) {
		fprintf(stderr, "%s: Send break.\n", CmdLine.PortPath);
		if (!TSerialPortSendBreak(&port)) {
			ret=0;
			goto out_close;
		}
	}

	if (!TM6KPanelInit(&m6k, CmdLine.DtL1L2)) {
		/* One mandatory function doesn't work. */
		ret=0;
		goto out_close;
	}
	buf_len=TM6KPanelSPrintfSize(&m6k,CmdLine.Format)
		+(sizeof(long) /* guard fence. */);
	buf=calloc(buf_len,sizeof(char));
	if (!buf) {
		fprintf(stderr,"%s: Can not allocate buffer.%s(%d).\n"
			,__func__
			,strerror(errno),errno
		);
		ret=0;
		goto out_close;
	}
	clear_queue=MAINB_CLEAR_QUEUE_FINISHED;
	if (CmdLine.Flags&COPT_FLAGS_CLEAR_QUEUE) {
		clear_queue=MAINB_CLEAR_QUEUE_START;
	}
	break_retry=0;
	pfds[0].fd=port.Fd;
	pfds[0].events=POLLIN|POLLERR|POLLHUP;
	pfds[1].fd=fileno(stdin);
	pfds[1].events=POLLIN|POLLERR|POLLHUP;
	while	(    (ProgContext.SignalIntr==0)
 		  && (ProgContext.SignalHup==0)
		  && (ProgContext.SignalTerm==0)
		) {
		unsigned char	c;
		int		timeout;

		ssize_t		read_bytes;

		pfds[0].revents=0;
		pfds[1].revents=0;
		timeout=CmdLine.Timeout;
		if (clear_queue!=MAINB_CLEAR_QUEUE_FINISHED) {
			timeout=MAINB_CLEAR_QUEUE_TIMEOUT;
		}
		poll_ret=poll(pfds,sizeof(pfds)/sizeof(pfds[0]),timeout);
		if (poll_ret==0) {
			/* Poll timeout. */
			if (clear_queue!=MAINB_CLEAR_QUEUE_FINISHED) {
				fprintf(stderr, "%s: Cleared queue. clear_queue=%d.\n", CmdLine.PortPath,clear_queue);
				clear_queue=MAINB_CLEAR_QUEUE_FINISHED;
				continue;
			}
			fprintf(stderr, "%s: Timeout.\n", CmdLine.PortPath);
			if (CmdLine.Flags&COPT_FLAGS_BREAK_AT_TIMEOUT) {
				fprintf(stderr, "%s: Reconfigure port.\n", CmdLine.PortPath);
				if (!TSerialPortConfigM6K(&port)) {
					break_retry++;
					continue;
				}
				fprintf(stderr, "%s: Send break.\n", CmdLine.PortPath);
				if (!TSerialPortSendBreak(&port)) {
					if (break_retry <= CmdLine.BreakRetry) {
						break_retry++;
						continue;
					}
				}
				break_retry=0;
			} else {
				ret=0;
				break;
			}
		}
		if (poll_ret<0) {
			if (errno==EINTR) {
				/* Interrupted. */
				fprintf(stderr, "%s: Catch signal.\n", ProgName);
				break;
			}
			fprintf(stderr, "%s: Poll failed. %s(%d).\n",CmdLine.PortPath
				,strerror(errno),errno
			);
			ret=0;
			break;
		}
		if (pfds[0].revents&POLLIN) {
			int	line_completed;

			read_bytes=read(port.Fd,&c,sizeof(c));
			if (read_bytes==0) {
				/* EOF, but why? */
				fprintf(stderr, "%s: Unexpected EOF.\n", CmdLine.PortPath);
				break;
			}
			if (read_bytes!=sizeof(c)) {
				fprintf(stderr, "%s: Read failed. %s(%d). read_bytes=%ld.\n"
					,CmdLine.PortPath
					,strerror(errno),errno
					,(long)(read_bytes)
				);
				ret=0;
				break;
			}
			if (clear_queue!=MAINB_CLEAR_QUEUE_FINISHED) {
				/* Clearling queue. */
				clear_queue++;
				continue;
			}
			if (TSerialPortDetectCountersError(&port)) {
				m6k.Flags|=TM6K_FLAGS_INVALID;
			}
			line_completed=TM6KPanelUpdateOne(&m6k,c);
			if (line_completed) {
				int	l;
				ssize_t	written_bytes;

				if ((CmdLine.Flags&COPT_FLAGS_INVALID_READ)==0) {
					/* Don't print invalid reads. */
					if (m6k.Flags&TM6K_FLAGS_INVALID) {
						/* Invalid read. */
						fprintf(stderr, "%s.UpdateOne: Invalid read.\n", CmdLine.PortPath);
						continue;
					}
				}
				/* May valid read. */
				l=TM6KPanelSPrintf(&m6k,CmdLine.Format,buf,buf_len);
				if (l==0) {
					fprintf(stderr, "%s: No output formed.\n"
						,__func__
					);
					ret=0;
					goto out_free;
				}
				if ((CmdLine.Flags&COPT_FLAGS_INVALID_READ)==0) {
					/* Don't print invalid reads. */
					if (m6k.Flags&TM6K_FLAGS_INVALID) {
						/* Invalid read. */
						fprintf(stderr, "%s.SPrintf: Invalid read.\n", CmdLine.PortPath);
						continue;
					}
				}
				*(buf+l-1)='\n';
				written_bytes=WriteRetry(fileno(stdout),buf,l);
				if (written_bytes!=l) {
					fprintf(stderr,"(stdout): Can not write stdout.\n");
					ret=0;
					goto out_free;
				}
			}
		}
		if (pfds[0].revents&(POLLERR|POLLHUP)) {
			fprintf(stderr,"%s: Hungup or error serial input. %s(%d).\n"
				,CmdLine.PortPath,strerror(errno),errno
				);
			ret=0;
			break;
		}
		if (pfds[1].revents&POLLIN) {
			/* user input. */
			ssize_t	read_bytes;
			char		*p;
			read_bytes=read(fileno(stdin),buf_stdin,sizeof(buf_stdin)-sizeof(long));
			if (read_bytes==0) {
				/* End of file. */
				break;
			}
			if (read_bytes<0) {
				fprintf(stderr,"(stdin): Read error. %s(%d).\n",strerror(errno),errno);
				ret=0;
				break;
			}
			/* Terminate with NUL. */
			*(buf_stdin+read_bytes)=0;
			p=buf_stdin;
			p=StrSkipSpace(p);
			switch (*p) {
				case 'q': {
					if (memcmp(p,"quit",4)==0) {
						fprintf(stderr, "%s: quit.\n",__func__);
						goto out_free;
					}
					break;
				}
				case 'e': {
					if (memcmp(p,"exit",4)==0) {
						fprintf(stderr,"%s: exit.\n",__func__);
						goto out_free;
					}
					break;
				}
				default:{ 
					/* do nothing. */
					break;
				}
			}
		}
		if (pfds[1].revents&(POLLERR|POLLHUP)) {
			fprintf(stderr,"(stdin): Hungup or error. %s(%d).\n",strerror(errno),errno);
			ret=0;
			break;
		}
	}
out_free:
	free(buf);
out_close:
	if (!TSerialPortClose(&port)) {
		ret=0;
	}
out:
	return(ret);
}

/*! Command line parser.
    @param copt command line parameters to store.
    @param argc same as main function's argc.
    @param argv same as main function's argv.
    @return int ==0: Parse failed, !=0: Parse success.
*/
int ParseCommandLine(TCommandLine *copt, int argc, char **argv)
{	int	opt;
	int	ret;

	ret=1 /* Success, true. */;
	while ((opt=getopt(argc, argv, "B:b:icf:k:t:D:Z:h"))!=-1) {
		switch(opt) {
			case 'B': {
				char	*p;
				p=StrSkipSpace(optarg);
				switch (*p) {
					case '1':
					case '+':
					case 'y':
						copt->Flags|=COPT_FLAGS_BREAK_AT_START;
						break;
					case '0':
					case '-':
					case 'n':
						copt->Flags&=~COPT_FLAGS_BREAK_AT_START;
						break;
					default:
						fprintf(stderr,"-B %s: Specify send break at start by y or n.\n",optarg);
						break;
				}
				break;
			}
			case 'b': {
				char	*p;;
				p=StrSkipSpace(optarg);
				switch (*p) {
					case '1':
					case '+':
					case 'y':
						copt->Flags|=COPT_FLAGS_BREAK_AT_TIMEOUT;
						break;
					case '0':
					case '-':
					case 'n':
						copt->Flags&=~COPT_FLAGS_BREAK_AT_TIMEOUT;
						break;
					default:
						fprintf(stderr,"-b %s: Specify send break at timeout by y or n.\n",optarg);
						break;
				}
				break;
			}
			case 'i': {
				/* Output invalid read. */
				copt->Flags|=COPT_FLAGS_INVALID_READ;
				break;
			}
			case 'c': {
				/* Clear queue. */
				copt->Flags|=COPT_FLAGS_CLEAR_QUEUE;
				break;
			}
			case 'f': {
				/* Format. */
				copt->Format=optarg;
				break;
			}
			case 'k': {
				/* Blank character. */
				copt->Blank=*optarg;
				break;
			}
			case 't': {
				/* Timeout. */
				char	*p;
				char	*p2;
				long	val;
				p=StrSkipSpace(optarg);
				val=strtol(p,&p2,0);
				if (p==p2) {
					fprintf(stderr,"-t %s: Specify timeout in milli seconds after -t.\n",optarg);
					copt->Flags|=COPT_FLAGS_HELP;
					ret=0 /* Failed, false. */;
					goto out;
				}
				copt->Timeout=val;
				break;
			}
			case 'D': {
				/* Delta time from End of line 1 to line 2. */
				char	*p;
				char	*p2;
				long	val;
				p=StrSkipSpace(optarg);
				val=strtol(p,&p2,0);
				if (p==p2) {
					fprintf(stderr,"-D %s: Specify deta time L1-L2 in milli seconds after -D.\n",optarg);
					copt->Flags|=COPT_FLAGS_HELP;
					ret=0 /* Failed, false. */;
					goto out;
				}
				if ((val<0L) || (val>=1000L)) {
					fprintf(stderr,"-D %s: Specify deta time in 0(internal-default) to 999.\n",optarg);
					copt->Flags|=COPT_FLAGS_HELP;
					ret=0 /* Failed, false. */;
					goto out;
				}
				copt->DtL1L2=val;
				break;
			}
			case 'Z': {
				/* Debugging option. */
				char	*p;
				p=StrSkipSpace(optarg);
				while (*p) {
					switch (*p) {
						case 's':
							copt->Flags|=COPT_FLAGS_SHOW_SETTINGS;
							break;
						case 'v':
							copt->Flags|=COPT_FLAGS_SHOW_VERSION;
							break;
						case 'l':
							copt->Flags|=COPT_FLAGS_DEBUG_LOCK;
							break;
						default:
							fprintf(stderr,"-X %s: Specify debug option.\n",optarg);
							break;
					}
					p++;
				}
				break;
			}
			case '?':
			case 'h': {
				/* Help. */
				copt->Flags|=COPT_FLAGS_HELP;
				goto out;
			}
			default: {
				fprintf(stderr,"-%c: Invalid option.\n",opt);
				copt->Flags|=COPT_FLAGS_HELP;
				goto out;
			}
		}
	}
	if ((copt->Flags&(COPT_FLAGS_HELP|COPT_FLAGS_SHOW_VERSION))==0) {
		if (optind>=argc) {
			fprintf(stderr,"Specify serial port path name.\n");
			ret=0;
			goto out;
		}
		copt->PortPath=argv[optind];
		if ((copt->PortPath==NULL) || (strlen(copt->PortPath)==0)) {
			fprintf(stderr,"Specify serial port path name.\n");
			ret=0;
			goto out;
		}
	}
out:
	return ret;
}

/*! Print help message.
*/
void ShowHelp(void)
{	fprintf(stdout,"AKIZUKI M-6000M tester serial output reader. version %s.\n"
		"%s"
		"Command line: [-B {y|n}] [-b {y|n}] [-c] [-D delta-time-L1-L2] [-f format] [-i] [-k char] [-t timeout] [-Z debug] [-h] serial_port_path\n"
		"-B{y|n}: send break at start.\n"
		"-b{y|n}: send break at timeout.\n"
		"-c: Clear queued chars at starting.\n"
		"-D milli-seconds: Delta time from 1st Line to 2nd Line (def: 40, range: 0..999).\n"
		"-f string: output Format.\n"
		"-i output invalid reads.\n"
		"-k character: blank character.\n"
		"-t milli-seconds: Timeout in milli seconds.\n"
		"-Z string: Debug. s: show settings, v: version string, l: phase lock state.\n"
		"-h: Help.\n"
		"Format specifier: \n"
		" %%a AC/DC\n"
		" %%v reading value without prefix.\n"
		" %%E reading value with engineering-exponential (e-9, e-6, .. ,e+6).\n"
		" %%e reading value with scientific-exponential.\n"
		" %%p prefix.\n"
		" %%u unit.\n"
		" %%h hold.\n"
		" %%m min/max.\n"
		" %%j judge.\n"
		" %%o over.\n"
		" %%n auto.\n"
		" %%b battery.\n"
		" %%T time reads in time_t.\n"
		" %%R real time reads in nano seconds.\n"
		" %%X raw reads in hex.\n"
		" %%Z raw reads.\n"
		,version_string
		,copyright_notice
	);
}

/*! Convert 0/1 to n/y.
    @param t ==0: means no, !=0: means yes.
    @return const char* "y" or "n".
*/
static const char *PlusMinus(int t)
{	if (t) {
		return "y";
	} else {
		return "n";
	}
}

/*! Show command line with settings.
*/
void ShowCommandLine(void)
{	unsigned long	flags;

	flags=CmdLine.Flags;
	fprintf(stdout,
		"SendBreakAtStart: -B %s\n"
		"SendBreakAtTimeout: -b %s\n"
		"ClearQueuedChars: %s\n"
		"DeltaTimeL1L2: -D %lu\n"
		"OutputFormat: -f \"%s\"\n"
		"BlankCharacter: -k '%c'\n"
		"TimeoutInMilliSeconds: -t %ld\n"
		"OutputInvalidReads: %s\n"
		"Debug: -Z %s%s\n"
		,PlusMinus(flags&COPT_FLAGS_BREAK_AT_START)
		,PlusMinus(flags&COPT_FLAGS_BREAK_AT_TIMEOUT)
		,(flags&COPT_FLAGS_CLEAR_QUEUE) ? "-c" :""
		,CmdLine.DtL1L2
		,CmdLine.Format
		,CmdLine.Blank
		,CmdLine.Timeout
		,(flags&COPT_FLAGS_INVALID_READ) ? "-i" : ""
		,(flags&COPT_FLAGS_SHOW_SETTINGS) ? "s" : ""
		,(flags&COPT_FLAGS_SHOW_VERSION) ? "v" : ""
	);
	fflush(stdout);
}

/*! Program main.
*/
int main(int argc, char **argv, char **env)
{	int		ret;
	int		parsed;
	unsigned long	flags;

	ret=0;
	ProgName=argv[0];

	if (!HandleSignals()) {
		fprintf(stderr,"%s: Can not handle signal.\n", ProgName);
		ret=2;
		goto out_unhandle;
	}
	parsed=ParseCommandLine(&CmdLine, argc, argv);
	flags=CmdLine.Flags;
	if (flags&COPT_FLAGS_SHOW_SETTINGS) {
		ShowCommandLine();
	}
	if (flags&COPT_FLAGS_SHOW_VERSION) {
		fprintf(stdout,"%s\n",version_string);
		ret=0;
		goto out_unhandle;
	}
	if (!parsed) {
		ShowHelp();
		ret=1;
		goto out_unhandle;
	}
	if (flags&COPT_FLAGS_HELP) {
		ShowHelp();
		ret=1;
		goto out_unhandle;
	}
	if (!MainB()) {
		ret=2;
	}
out_unhandle:
	UnHandleSignals();
	return(ret);
}

